"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const anchor_1 = require("@coral-xyz/anchor");
const web3_js_1 = require("@solana/web3.js");
const dotenv = __importStar(require("dotenv"));
const node_fetch_1 = __importDefault(require("node-fetch"));
const client_1 = require("../src/client");
const constants_1 = require("../src/constants");
const utils_1 = require("../src/utils");
dotenv.config();
const CLUSTER_URL = process.env.CLUSTER_URL_OVERRIDE || process.env.MB_CLUSTER_URL;
const GROUP_PK = process.env.GROUP_PK || '78b8f4cGCwmZ9ysPFMWLaLTkkaYnUjwMJYStWe5RTSSX';
const CLUSTER = process.env.CLUSTER_OVERRIDE || 'mainnet-beta';
async function main() {
    const options = anchor_1.AnchorProvider.defaultOptions();
    const connection = new web3_js_1.Connection(CLUSTER_URL, options);
    const wallet = new anchor_1.Wallet(new web3_js_1.Keypair());
    const provider = new anchor_1.AnchorProvider(connection, wallet, options);
    const client = client_1.MangoClient.connect(provider, CLUSTER, constants_1.MANGO_V4_ID[CLUSTER], {
        idsSource: 'get-program-accounts',
    });
    const group = await client.getGroup(new web3_js_1.PublicKey(GROUP_PK));
    const mangoAccounts = await client.getAllMangoAccounts(group, true);
    const largeMangoAccounts = mangoAccounts
        .sort((a, b) => b.getEquity(group).cmp(a.getEquity(group)))
        .filter((a) => (0, utils_1.toUiDecimalsForQuote)(a.getEquity(group)) > 1000);
    console.log(`Table 1: On-chain vs Offchain Pnl`);
    console.log(`${'Account'.padStart(48)} ${'Equity'.padStart(16)} ${'Net deposits'.padStart(16)} ${'On-chain PnL'.padStart(16)} ${'Off-chain PnL'.padStart(16)} ${'Off-chain PnL Date'.padStart(16)} ${'Diff'.padStart(16)}`);
    for (const a of largeMangoAccounts) {
        const url = `https://api.mngo.cloud/data/v4/stats/performance_account?mango-account=${a.publicKey}&start-date=2023-04-29`;
        const resp = await (0, node_fetch_1.default)(url);
        const data = await resp.json();
        const keys = Object.keys(data).sort((a, b) => new Date(a).getTime() - new Date(b).getTime());
        const latestEntry = keys.length - 1;
        const offChainPnl = data[keys[latestEntry]].pnl;
        const onChain = new anchor_1.BN(a.getEquity(group).floor().toNumber()).sub(a.netDeposits);
        const diff = (0, utils_1.toUiDecimalsForQuote)(onChain) - offChainPnl;
        console.log(`${a.publicKey.toString().padStart(48)} ${(0, utils_1.toUiDecimalsForQuote)(a.getEquity(group))
            .toLocaleString()
            .padStart(15)}$ ${(0, utils_1.toUiDecimalsForQuote)(a.netDeposits)
            .toLocaleString()
            .padStart(15)}$ ${(0, utils_1.toUiDecimalsForQuote)(onChain)
            .toLocaleString()
            .padStart(15)}$ ${offChainPnl.toLocaleString().padStart(15)}$ (${keys[keys.length - 1]}) ${diff.toLocaleString().padStart(15)}$`);
    }
    console.log();
    console.log(`Table 2: On-chain vs Offchain Cumulating funding`);
    console.log(`${'Account'.padStart(48)} ${'BTC On-chain'.padStart(11)} ${'BTC Off-chain'.padStart(11)} ${'diff'.padStart(11)}`);
    function funding(group, marketIndex, mangoAccount, data) {
        const pps = mangoAccount
            .perpActive()
            .filter((a) => a.marketIndex === marketIndex);
        if (pps.length == 0) {
            return [-1, -1];
        }
        const pp = pps[0];
        const onChain = (0, utils_1.toUiDecimalsForQuote)(pp.cumulativeLongFunding - pp.cumulativeShortFunding);
        const offChainEntry = data[group.getPerpMarketByMarketIndex(marketIndex).name];
        if (!offChainEntry) {
            return [onChain, -1];
        }
        const offChain = offChainEntry['long_funding'] + offChainEntry['short_funding'];
        return [onChain, offChain];
    }
    for (const a of largeMangoAccounts) {
        const resp = await (0, node_fetch_1.default)(`https://api.mngo.cloud/data/v4/stats/funding-account-total?mango-account=${a.publicKey}`);
        const data = await resp.json();
        const btc = funding(group, 0, a, data);
        // const sol = funding(group, 2 as PerpMarketIndex, a, data);
        // const eth = funding(group, 3 as PerpMarketIndex, a, data);
        if (btc[0] != -1 && btc[1] != -1)
            console.log(`${a.publicKey.toString().padStart(48)} ${btc[0]
                .toLocaleString()
                .padStart(10)}$ ${btc[1].toLocaleString().padStart(10)}$ ${(btc[0] - btc[1])
                .toLocaleString()
                .padStart(10)}$`);
    }
}
try {
    main();
}
catch (error) {
    console.log(error);
}
